@@ -25,6 +25,10 @@
 #include "gclue-service-client.h"
 #include "gclue-service-location.h"
 #include "gclue-locator.h"
+#include "public-api/gclue-enum-types.h"
+#include "gclue-config.h"
+
+#define DEFAULT_ACCURACY_LEVEL GCLUE_ACCURACY_LEVEL_CITY
 
 static void
 gclue_service_client_client_iface_init (GClueClientIface *iface);
@@ -41,9 +45,10 @@ G_DEFINE_TYPE_WITH_CODE (GClueServiceClient,
 
 struct _GClueServiceClientPrivate
 {
-        const char *peer;
+        GClueClientInfo *client_info;
         const char *path;
         GDBusConnection *connection;
+        GClueAgent *agent_proxy;
 
         GClueServiceLocation *location;
         GClueServiceLocation *prev_location;
@@ -54,28 +59,21 @@ struct _GClueServiceClientPrivate
         /* Number of times location has been updated */
         guint locations_updated;
 
-        guint watch_id;
-        gulong location_change_id;
+        gboolean agent_stopped; /* Agent stopped client, not the app */
 };
 
 enum
 {
         PROP_0,
-        PROP_PEER,
+        PROP_CLIENT_INFO,
         PROP_PATH,
         PROP_CONNECTION,
+        PROP_AGENT_PROXY,
         LAST_PROP
 };
 
 static GParamSpec *gParamSpecs[LAST_PROP];
 
-enum {
-        PEER_VANISHED,
-        SIGNAL_LAST
-};
-
-static guint signals[SIGNAL_LAST];
-
 static char *
 next_location_path (GClueServiceClient *client)
 {
@@ -100,11 +98,13 @@ emit_location_updated (GClueServiceClient *client,
 {
         GClueServiceClientPrivate *priv = client->priv;
         GVariant *variant;
+        const char *peer;
 
         variant = g_variant_new ("(oo)", old, new);
+        peer = gclue_client_info_get_bus_name (priv->client_info);
 
         return g_dbus_connection_emit_signal (priv->connection,
-                                              priv->peer,
+                                              peer,
                                               priv->path,
                                               "org.freedesktop.GeoClue2.Client",
                                               "LocationUpdated",
@@ -145,13 +145,15 @@ on_locator_location_changed (GObject    *gobject,
 {
         GClueServiceClient *client = GCLUE_SERVICE_CLIENT (user_data);
         GClueServiceClientPrivate *priv = client->priv;
-        GClueLocator *locator = GCLUE_LOCATOR (gobject);
+        GClueLocationSource *locator = GCLUE_LOCATION_SOURCE (gobject);
         GeocodeLocation *location_info;
         char *path = NULL;
         const char *prev_path;
         GError *error = NULL;
 
-        location_info = gclue_locator_get_location (locator);
+        location_info = gclue_location_source_get_location (locator);
+        if (location_info == NULL)
+                return; /* No location found yet */
 
         if (priv->location != NULL && below_threshold (client, location_info)) {
                 g_debug ("Updating location, below threshold");
@@ -165,7 +167,7 @@ on_locator_location_changed (GObject    *gobject,
         priv->prev_location = priv->location;
 
         path = next_location_path (client);
-        priv->location = gclue_service_location_new (priv->peer,
+        priv->location = gclue_service_location_new (priv->client_info,
                                                      path,
                                                      priv->connection,
                                                      location_info,
@@ -191,6 +193,77 @@ out:
         g_free (path);
 }
 
+static void
+start_client (GClueServiceClient *client, GClueAccuracyLevel accuracy_level)
+{
+        GClueServiceClientPrivate *priv = client->priv;
+
+        gclue_client_set_active (GCLUE_CLIENT (client), TRUE);
+        priv->locator = gclue_locator_new (accuracy_level);
+        g_signal_connect (priv->locator,
+                          "notify::location",
+                          G_CALLBACK (on_locator_location_changed),
+                          client);
+
+        gclue_location_source_start (GCLUE_LOCATION_SOURCE (priv->locator));
+}
+
+static void
+stop_client (GClueServiceClient *client)
+{
+        g_clear_object (&client->priv->locator);
+        gclue_client_set_active (GCLUE_CLIENT (client), FALSE);
+}
+
+static void
+on_agent_props_changed (GDBusProxy *agent_proxy,
+                        GVariant   *changed_properties,
+                        GStrv       invalidated_properties,
+                        gpointer    user_data)
+{
+        GClueServiceClient *client = GCLUE_SERVICE_CLIENT (user_data);
+        GVariantIter *iter;
+        GVariant *value;
+        gchar *key;
+        
+        if (g_variant_n_children (changed_properties) < 0)
+                return;
+
+        g_variant_get (changed_properties, "a{sv}", &iter);
+        while (g_variant_iter_loop (iter, "{&sv}", &key, &value)) {
+                GClueAccuracyLevel max_accuracy;
+                const char *id;
+
+                if (strcmp (key, "MaxAccuracyLevel") != 0)
+                        continue;
+
+                id = gclue_client_get_desktop_id (GCLUE_CLIENT (client));
+                max_accuracy = g_variant_get_uint32 (value);
+                /* FIXME: We should be handling all values of max accuracy
+                 *        level here, not just 0 and non-0.
+                 */
+                if (max_accuracy != 0 && client->priv->agent_stopped) {
+                        GClueAccuracyLevel accuracy;
+
+                        client->priv->agent_stopped = FALSE;
+                        accuracy = gclue_client_get_requested_accuracy_level
+                                (GCLUE_CLIENT (client));
+                        accuracy = CLAMP (accuracy, 0, max_accuracy);
+
+                        start_client (client, accuracy);
+                        g_debug ("Re-started '%s'.", id);
+                } else if (max_accuracy == 0 &&
+                           gclue_client_get_active (GCLUE_CLIENT (client))) {
+                        stop_client (client);
+                        client->priv->agent_stopped = TRUE;
+                        g_debug ("Stopped '%s'.", id);
+                }
+
+                break;
+        }
+        g_variant_iter_free (iter);
+}
+
 typedef struct
 {
         GClueServiceClient *client;
@@ -198,63 +271,122 @@ typedef struct
 } StartData;
 
 static void
-on_start_ready (GObject      *source_object,
-                GAsyncResult *res,
-                gpointer      user_data)
+start_data_free (StartData *data)
 {
-        StartData *data = (StartData *) user_data;
-        GClueLocator *locator = GCLUE_LOCATOR (source_object);
-        GError *error = NULL;
-
-        if (!gclue_locator_start_finish (locator, res, &error))
-                goto error_out;
-
-        gclue_client_complete_start (GCLUE_CLIENT (data->client),
-                                     data->invocation);
-        goto out;
-
-error_out:
-        g_dbus_method_invocation_return_error (data->invocation,
-                                               G_DBUS_ERROR,
-                                               G_DBUS_ERROR_FAILED,
-                                               "Failed to start: %s",
-                                               error->message);
-        g_error_free (error);
-
-out:
         g_object_unref (data->client);
         g_object_unref (data->invocation);
         g_slice_free (StartData, data);
 }
 
+static void
+complete_start (StartData *data, GClueAccuracyLevel accuracy_level)
+{
+        start_client (data->client, accuracy_level);
+
+        gclue_client_complete_start (GCLUE_CLIENT (data->client),
+                                     data->invocation);
+        g_debug ("'%s' started.",
+                 gclue_client_get_desktop_id (GCLUE_CLIENT (data->client)));
+        start_data_free (data);
+}
+
+static void
+on_authorize_app_ready (GObject      *source_object,
+                        GAsyncResult *res,
+                        gpointer      user_data)
+{
+        StartData *data = (StartData *) user_data;
+        GError *error = NULL;
+        gboolean authorized = FALSE;
+        GClueAccuracyLevel accuracy_level;
+
+        accuracy_level = gclue_client_get_requested_accuracy_level
+                                (GCLUE_CLIENT (data->client));
+        if (!gclue_agent_call_authorize_app_finish (GCLUE_AGENT (source_object),
+                                                    &authorized,
+                                                    &accuracy_level,
+                                                    res,
+                                                    &error))
+                goto error_out;
+
+        if (!authorized) {
+                g_set_error_literal (&error,
+                                     G_DBUS_ERROR,
+                                     G_DBUS_ERROR_ACCESS_DENIED,
+                                     "Access denied");
+                goto error_out;
+        }
+
+        complete_start (data, accuracy_level);
+
+        return;
+
+error_out:
+        g_dbus_method_invocation_take_error (data->invocation, error);
+        start_data_free (data);
+}
+
 static gboolean
 gclue_service_client_handle_start (GClueClient           *client,
                                    GDBusMethodInvocation *invocation)
 {
         GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (client)->priv;
+        GClueConfig *config;
         StartData *data;
+        const char *desktop_id;
+        GClueAccuracyLevel accuracy_level, max_accuracy;
 
-        if (priv->location_change_id)
+        if (priv->locator != NULL)
                 /* Already started */
                 return TRUE;
 
-        /* TODO: We need some kind of mechanism to ask user if location info can
-         *       be shared with peer app.
-         */
-
-        priv->location_change_id =
-                g_signal_connect (priv->locator,
-                                  "notify::location",
-                                  G_CALLBACK (on_locator_location_changed),
-                                  client);
+        desktop_id = gclue_client_get_desktop_id (client);
+        if (desktop_id == NULL) {
+                g_dbus_method_invocation_return_error (invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                       "'DesktopId' property must be set");
+                return TRUE;
+        }
 
         data = g_slice_new (StartData);
         data->client = g_object_ref (client);
         data->invocation =  g_object_ref (invocation);
-        gclue_locator_start (priv->locator,
-                             NULL,
-                             on_start_ready,
-                             data);
+
+        config = gclue_config_get_singleton ();
+        accuracy_level = gclue_client_get_requested_accuracy_level (client);
+
+        /* No agent == No authorization needed */
+        if (priv->agent_proxy == NULL ||
+            gclue_config_is_app_allowed (config,
+                                         desktop_id,
+                                         priv->client_info)) {
+                complete_start (data, accuracy_level);
+
+                return TRUE;
+        }
+
+        max_accuracy = gclue_agent_get_max_accuracy_level (priv->agent_proxy);
+        if (max_accuracy == 0) {
+                guint32 uid = gclue_client_info_get_user_id (priv->client_info);
+
+                g_dbus_method_invocation_return_error (invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                       "Geolocation disabled for"
+                                                       " UID %u",
+                                                       uid);
+                start_data_free (data);
+                return TRUE;
+        }
+        accuracy_level = CLAMP (accuracy_level, 0, max_accuracy);
+
+        gclue_agent_call_authorize_app (priv->agent_proxy,
+                                        desktop_id,
+                                        accuracy_level,
+                                        NULL,
+                                        on_authorize_app_ready,
+                                        data);
 
         return TRUE;
 }
@@ -263,15 +395,9 @@ static gboolean
 gclue_service_client_handle_stop (GClueClient           *client,
                                   GDBusMethodInvocation *invocation)
 {
-        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (client)->priv;
-
-        if (priv->location_change_id != 0) {
-                g_signal_handler_disconnect (priv->locator,
-                                             priv->location_change_id);
-                priv->location_change_id = 0;
-        }
-
+        stop_client (GCLUE_SERVICE_CLIENT (client));
         gclue_client_complete_stop (client, invocation);
+        g_debug ("'%s' stopped.", gclue_client_get_desktop_id (client));
 
         return TRUE;
 }
@@ -281,18 +407,13 @@ gclue_service_client_finalize (GObject *object)
 {
         GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (object)->priv;
 
-        if (priv->location_change_id != 0) {
-                g_signal_handler_disconnect (priv->locator,
-                                             priv->location_change_id);
-                priv->location_change_id = 0;
-        }
-
-        g_clear_pointer (&priv->peer, g_free);
         g_clear_pointer (&priv->path, g_free);
         g_clear_object (&priv->connection);
+        g_clear_object (&priv->agent_proxy);
         g_clear_object (&priv->locator);
         g_clear_object (&priv->location);
         g_clear_object (&priv->prev_location);
+        g_clear_object (&priv->client_info);
 
         /* Chain up to the parent class */
         G_OBJECT_CLASS (gclue_service_client_parent_class)->finalize (object);
@@ -307,8 +428,8 @@ gclue_service_client_get_property (GObject    *object,
         GClueServiceClient *client = GCLUE_SERVICE_CLIENT (object);
 
         switch (prop_id) {
-        case PROP_PEER:
-                g_value_set_string (value, client->priv->peer);
+        case PROP_CLIENT_INFO:
+                g_value_set_object (value, client->priv->client_info);
                 break;
 
         case PROP_PATH:
@@ -319,6 +440,10 @@ gclue_service_client_get_property (GObject    *object,
                 g_value_set_object (value, client->priv->connection);
                 break;
 
+        case PROP_AGENT_PROXY:
+                g_value_set_object (value, client->priv->agent_proxy);
+                break;
+
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         }
@@ -333,8 +458,8 @@ gclue_service_client_set_property (GObject      *object,
         GClueServiceClient *client = GCLUE_SERVICE_CLIENT (object);
 
         switch (prop_id) {
-        case PROP_PEER:
-                client->priv->peer = g_value_dup_string (value);
+        case PROP_CLIENT_INFO:
+                client->priv->client_info = g_value_dup_object (value);
                 break;
 
         case PROP_PATH:
@@ -345,6 +470,14 @@ gclue_service_client_set_property (GObject      *object,
                 client->priv->connection = g_value_dup_object (value);
                 break;
 
+        case PROP_AGENT_PROXY:
+                client->priv->agent_proxy = g_value_dup_object (value);
+                g_signal_connect (client->priv->agent_proxy,
+                                  "g-properties-changed",
+                                  G_CALLBACK (on_agent_props_changed),
+                                  object);
+                break;
+
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         }
@@ -364,7 +497,7 @@ gclue_service_client_handle_method_call (GDBusConnection       *connection,
         GDBusInterfaceSkeletonClass *skeleton_class;
         GDBusInterfaceVTable *skeleton_vtable;
 
-        if (strcmp (sender, priv->peer) != 0) {
+        if (!gclue_client_info_check_bus_name (priv->client_info, sender)) {
                 g_dbus_method_invocation_return_error (invocation,
                                                        G_DBUS_ERROR,
                                                        G_DBUS_ERROR_ACCESS_DENIED,
@@ -397,7 +530,7 @@ gclue_service_client_handle_get_property (GDBusConnection *connection,
         GDBusInterfaceSkeletonClass *skeleton_class;
         GDBusInterfaceVTable *skeleton_vtable;
 
-        if (strcmp (sender, priv->peer) != 0) {
+        if (!gclue_client_info_check_bus_name (priv->client_info, sender)) {
                 g_set_error (error,
                              G_DBUS_ERROR,
                              G_DBUS_ERROR_ACCESS_DENIED,
@@ -432,7 +565,7 @@ gclue_service_client_handle_set_property (GDBusConnection *connection,
         GDBusInterfaceVTable *skeleton_vtable;
         gboolean ret;
 
-        if (strcmp (sender, priv->peer) != 0) {
+        if (!gclue_client_info_check_bus_name (priv->client_info, sender)) {
                 g_set_error (error,
                              G_DBUS_ERROR,
                              G_DBUS_ERROR_ACCESS_DENIED,
@@ -488,15 +621,15 @@ gclue_service_client_class_init (GClueServiceClientClass *klass)
 
         g_type_class_add_private (object_class, sizeof (GClueServiceClientPrivate));
 
-        gParamSpecs[PROP_PEER] = g_param_spec_string ("peer",
-                                                      "Peer",
-                                                      "Bus name of client app",
-                                                      NULL,
-                                                      G_PARAM_READWRITE |
-                                                      G_PARAM_CONSTRUCT_ONLY);
+        gParamSpecs[PROP_CLIENT_INFO] = g_param_spec_object ("client-info",
+                                                             "ClientInfo",
+                                                             "Information on client",
+                                                             GCLUE_TYPE_CLIENT_INFO,
+                                                             G_PARAM_READWRITE |
+                                                             G_PARAM_CONSTRUCT_ONLY);
         g_object_class_install_property (object_class,
-                                         PROP_PEER,
-                                         gParamSpecs[PROP_PEER]);
+                                         PROP_CLIENT_INFO,
+                                         gParamSpecs[PROP_CLIENT_INFO]);
 
         gParamSpecs[PROP_PATH] = g_param_spec_string ("path",
                                                       "Path",
@@ -518,18 +651,15 @@ gclue_service_client_class_init (GClueServiceClientClass *klass)
                                          PROP_CONNECTION,
                                          gParamSpecs[PROP_CONNECTION]);
 
-        signals[PEER_VANISHED] =
-                g_signal_new ("peer-vanished",
-                              GCLUE_TYPE_SERVICE_CLIENT,
-                              G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GClueServiceClientClass,
-                                               peer_vanished),
-                              NULL,
-                              NULL,
-                              g_cclosure_marshal_VOID__VOID,
-                              G_TYPE_NONE,
-                              0,
-                              G_TYPE_NONE);
+        gParamSpecs[PROP_AGENT_PROXY] = g_param_spec_object ("agent-proxy",
+                                                             "AgentProxy",
+                                                             "Proxy to app authorization agent",
+                                                             GCLUE_TYPE_AGENT_PROXY,
+                                                             G_PARAM_READWRITE |
+                                                             G_PARAM_CONSTRUCT);
+        g_object_class_install_property (object_class,
+                                         PROP_AGENT_PROXY,
+                                         gParamSpecs[PROP_AGENT_PROXY]);
 }
 
 static void
@@ -539,25 +669,11 @@ gclue_service_client_client_iface_init (GClueClientIface *iface)
         iface->handle_stop = gclue_service_client_handle_stop;
 }
 
-static void
-on_name_vanished (GDBusConnection *connection,
-                  const gchar     *name,
-                  gpointer         user_data)
-{
-        g_signal_emit (GCLUE_SERVICE_CLIENT (user_data),
-                       signals[PEER_VANISHED],
-                       0);
-
-        g_object_unref (G_OBJECT (user_data));
-}
-
 static gboolean
 gclue_service_client_initable_init (GInitable    *initable,
                                     GCancellable *cancellable,
                                     GError      **error)
 {
-        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (initable)->priv;
-
         if (!g_dbus_interface_skeleton_export
                                 (G_DBUS_INTERFACE_SKELETON (initable),
                                  GCLUE_SERVICE_CLIENT (initable)->priv->connection,
@@ -565,13 +681,6 @@ gclue_service_client_initable_init (GInitable    *initable,
                                  error))
                 return FALSE;
 
-        priv->watch_id = g_bus_watch_name_on_connection (priv->connection,
-                                                         priv->peer,
-                                                         G_BUS_NAME_WATCHER_FLAGS_NONE, 
-                                                         NULL,
-                                                         on_name_vanished,
-                                                         g_object_ref (initable),
-                                                         g_object_unref);
         return TRUE;
 }
 
@@ -587,22 +696,24 @@ gclue_service_client_init (GClueServiceClient *client)
         client->priv = G_TYPE_INSTANCE_GET_PRIVATE (client,
                                                     GCLUE_TYPE_SERVICE_CLIENT,
                                                     GClueServiceClientPrivate);
-
-        client->priv->locator = gclue_locator_new ();
+        gclue_client_set_requested_accuracy_level (GCLUE_CLIENT (client),
+                                                   DEFAULT_ACCURACY_LEVEL);
 }
 
 GClueServiceClient *
-gclue_service_client_new (const char      *peer,
+gclue_service_client_new (GClueClientInfo *info,
                           const char      *path,
                           GDBusConnection *connection,
+                          GClueAgent      *agent_proxy,
                           GError         **error)
 {
         return g_initable_new (GCLUE_TYPE_SERVICE_CLIENT,
                                NULL,
                                error,
-                               "peer", peer,
+                               "client-info", info,
                                "path", path,
                                "connection", connection,
+                               "agent-proxy", agent_proxy,
                                NULL);
 }
 
@@ -614,10 +725,10 @@ gclue_service_client_get_path (GClueServiceClient *client)
         return client->priv->path;
 }
 
-const gchar *
-gclue_service_client_get_peer (GClueServiceClient *client)
+GClueClientInfo *
+gclue_service_client_get_client_info (GClueServiceClient *client)
 {
         g_return_val_if_fail (GCLUE_IS_SERVICE_CLIENT(client), NULL);
 
-        return client->priv->peer;
+        return client->priv->client_info;
 }
